import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:isolate';
import 'dart:typed_data';
import 'package:crypto/crypto.dart';
import 'package:image_compression/image_compression.dart';

import 'http/http_connector.dart';

import 'moderator/OutgoingMessage.dart';
import 'moderator/OutgoingMessageCache.dart';
import 'moderator/moderatorCore.dart';
import 'moderator/OmniRemembers.dart';
import 'moderator/ogParse.dart';

import 'disk.dart';
export 'moderator/moderatorCore.dart';
export 'disk.dart';
import 'package:hex/hex.dart';
import 'interop/xxhash/lib/fast_hash.dart';
import 'package:blurhash_dart/blurhash_dart.dart';
import 'package:image/image.dart' as img;
import 'ipfsQueue.dart';
export 'disk.dart';
import 'package:mime/mime.dart';

class world {
  httpInterzone? _io;
  interzoneDisk? _disk;
  OmniRemembers _omni = OmniRemembers();
  ModeratorEntrySet data = ModeratorEntrySet();
  ModeratorCapsule? _latestCapsule;

  xxCache _xxCache = xxCache();
  Future<bool> satisfyXX(String xx, Uint8List data) =>
      _xxCache.satisfy(xx, data, false);
  OutgoingMessageCache _ogMessages = OutgoingMessageCache();
  String _server = '';
  String _siteName = '';
  String get siteName => _siteName;
  String get baseUrl => _server;

  String _nicke = '';
  String _avatare = '🐷';
  String get currentNick => _nicke;
  String get currentAvatar => _avatare;
  String appSplashImage = '';
  bool _dirty = false;
  IpfsQueue _ipfsQ = IpfsQueue();
  bool _terminated = false;
  int dataIn = 0;
  int dataOut = 0;
  bool get _online => _io != null ? _io!.online : false;
  List<OutgoingMessage> get currentPending => _ogMessages.currentPendingItems;
  void setBaseUrl(String url) => _server = url;

  void setOnline() {
    _io?.setOnline;
    _ipfsQ.setOnline;
    _xxCache.setOnline();
  }

  void setOffline() {
    _io?.setOffline;
    _ipfsQ.setOffline;
    _xxCache.setOffline();
  }

  void unDirty() {
    //call this from outside only
    _dirty = false;
  }

  bool get isPrimed => data.isPrimed;
  bool get isDirty => _dirty;
  world();

  xxCache get xxC => _xxCache;
  Uint8List get xxWants => _xxCache.wantsBinary;
  //action handlers
  dynamic messageActionSignalHandler;
  Set<String> _mutedAuthors = {};
  Set<int> _deletedEntries = {};
  void terminate() {
    _terminated = true;
  }

  bool _housekeeping = false;
  void housekeep() {
    if (_housekeeping) return;
    _housekeeping = true;
    if (_latestCapsule == null) {
      print('housekeep latestCapsule empty, try login');
      login.then((value) {
        print('housekeep latestCapsule empty, login result');
        if (value != null) {
          print(value.length);
        } else {
          print('no logn');
        }
      });
    }

    //saving stuff_xxCache.resetIPFSpullQueue();
    _xxCache.housekeep();
    //clear borked outgoing messages
    _ogMessages.housekeep();
    //pretend we dont know anything
    _latelySeenp4p.clear();
    _housekeeping = false;
  }

  int get now => DateTime.now().millisecondsSinceEpoch;
  void ping() {
    refreshViaHTTP();
  }

  void defocus() {
    //TODO
  }

  int get dataPulledDisk => _disk!.dataPulledDisk;
  int get dataPushedDisk => _disk!.dataPushedDisk;

  ModeratorEntrySet get messages {
    if (_mutedAuthors.isEmpty && _deletedEntries.isEmpty) return data;
    ModeratorEntrySet _l = ModeratorEntrySet();
    data.all.forEach((element) {
      if (_mutedAuthors.contains(element.author)) return;
      if (_deletedEntries.contains(element.keyHint)) return;
      element.flags.isSeen = _omni.isSeen(element.hint);
      element.flags.isLiked = _omni.isLiked(element.hint);
      _l.add(element);
    });
    return _l;
  }

  ModeratorEntrySet get LatestThreads {
    //if (_mutedAuthors.isEmpty && _deletedEntries.isEmpty) return data;
    ModeratorEntrySet _l = ModeratorEntrySet();
    data.latestThreads.forEach((element) {
      if (!element.flags.isPost) return;
      if (_mutedAuthors.contains(element.author)) return;
      if (_deletedEntries.contains(element.keyHint)) return;
      element.flags.isSeen = _omni.isSeen(element.hint);
      element.flags.isLiked = _omni.isLiked(element.hint);
      //patch user perception
      //if (element.views == 0 && element.flags.isSeen) element.views = 1;
      if (element.likes == 0 && element.flags.isLiked) element.likes = 1;
      _l.add(element);
    });
    return _l;
  }

  List<ModeratorEntry> _currentPendingAsModeratorEntries = [];
  List<ModeratorEntry> get currentPendingAsModeratorEntries {
    List<ModeratorEntry> _l = [];

    for (OutgoingMessage i in currentPending) {
      //skip non post reply
      if (i.type != 0 && i.type != 1) continue;
      Uint8List _body = utf8.encode(i.text) as Uint8List;
      Uint8List _sl = Uint8List(0);
      int bl = 0;
      int pp = 0;
      int keyHint = i.timestamp;
      String colour = '';

      Uint8List _xx = utf8.encode(i.xxHash) as Uint8List;
      Uint8List _atl = Uint8List(8)
        ..buffer.asByteData().setInt64(0, i.attachementLength, Endian.big);
      Uint8List attachmentLink = Uint8List(16);
      int sds = 0;
      for (var i in _atl) {
        attachmentLink[sds++] = i;
      }
      for (var i in HEX.decode(i.xxHash)) {
        attachmentLink[sds++] = i;
      }

      //Uint8List attachmentLink = Uint8List(8)
      //  ..buffer.asByteData().setInt64(0, i.xxHash, Endian.big);

      Uint8List _eflags = Uint8List(2);
      ModeratorEntryFlags flags = ModeratorEntryFlags(_eflags);
      flags.moderate = true;
      flags.attachements = i.hasAttachment;
      if (i.type == 0) flags.isPost = true;
      if (i.type == 1) flags.isReply = true;
      //keyhint -1 filter out pending on ui level
      final me = ModeratorEntry(
          i.type,
          i.nick,
          _body,
          _sl,
          keyHint,
          bl,
          pp,
          colour,
          i.avatar,
          i.nick,
          attachmentLink,
          flags,
          i.timestamp,
          0,
          0,
          0,
          0);
      me.isPending = true;
      _l.add(me);
    }
    _currentPendingAsModeratorEntries = _l;
    return _l;
  }

  ModeratorEntrySet get LatestThreadsWithOutgoingItems {
    //if (_mutedAuthors.isEmpty && _deletedEntries.isEmpty) return data;
    ModeratorEntrySet _l = data;
    currentPendingAsModeratorEntries;
    for (ModeratorEntry i in currentPendingAsModeratorEntries) {
      _l.add(i);
    }
    List<ModeratorEntry> _lp = [];
    _l.all.forEach((element) {
      if (!element.flags.isPost) return;
      if (_mutedAuthors.contains(element.author)) return;
      if (_deletedEntries.contains(element.keyHint)) return;
      element.flags.isSeen = _omni.isSeen(element.hint);
      element.flags.isLiked = _omni.isLiked(element.hint);
      //patch user perception
      //if (element.views == 0 && element.flags.isSeen) element.views = 1;
      if (element.likes == 0 && element.flags.isLiked) element.likes = 1;
      _l.add(element);
    });
    return _l;
  }

  List<ModeratorEntry> threadByShortLink(String shortLink) {
    List<ModeratorEntry> _ll = [];
    final ll2 = data.threadByShortLink(shortLink);
    if (ll2.isEmpty) return _ll;
    for (ModeratorEntry i in _currentPendingAsModeratorEntries) {
      if (i.bl == ll2.first.keyHint) {
        _ll.add(i);
      }
    }

    bool pkc = false;
    for (var element in ll2) {
      if (_deletedEntries.contains(element.keyHint)) continue;
      if (element.flags.isPost && !pkc) {
        pkc = true;
        bool newa = false;
        List<ModeratorEntry> _lll = [];
        ll2.forEach((ce) {
          if (_deletedEntries.contains(ce.keyHint)) return;
          if (ce.flags.isReply) {
            ce.flags.isSeen = _omni.isSeen(ce.hint);
            ce.flags.isLiked = _omni.isLiked(ce.hint);
            //any single item unseen makes parent unseen
            if (!ce.flags.isSeen) newa = true;
            _lll.add(ce);
          }
        });
        if (newa) {
          element.flags.isSeen = false;
        } else {
          element.flags.isSeen = _omni.isSeen(element.hint);
        }
        _ll.add(element);
        if (_lll.isNotEmpty) _ll.addAll(_lll);
        break;
      }
    }

    return _ll;
  }

  List<ModeratorEntry> unreadThreadByShortLink(String shortLink) {
    //OMNI here to track seen unseen / seen
    List<ModeratorEntry> _ll = [];
    for (var element in data.threadByShortLink(shortLink)) {
      if (element.shortLink == shortLink) {
        bool newa = false;
        List<ModeratorEntry> _lll = [];
        data.childrenOf(element.hint).forEach((ce) {
          if (ce.flags.isReply) {
            ce.flags.isSeen = _omni.isSeen(ce.hint);
            //any single item unseen makes parent unseen
            if (!ce.flags.isSeen) newa = true;
            _lll.add(ce);
          }
        });
        if (newa) {
          element.flags.isSeen = false;
        } else {
          element.flags.isSeen = _omni.isSeen(element.hint);
        }
        _ll.add(element);
        if (_lll.isNotEmpty) _ll.addAll(_lll);
      }
    }
    return _ll;
  }

  bool isSeenOmni(String shortLink) {
    ModeratorEntry? m = data.read(shortLink);
    if (m == null) return false;
    return _omni.isSeen(m.hint);
  }

  bool isLikedOmni(String shortLink) {
    ModeratorEntry? m = data.read(shortLink);
    if (m == null) return false;
    return _omni.isLiked(m.hint);
  }

  bool markSeenOmni(String shortLink) {
    //set this entry as seen
    ModeratorEntry? m = data.read(shortLink);
    if (m == null) return false;
    if (_omni.isSeen(m.hint)) return true;
    //refresh for cache
    _dirty = true;
    return _omni.markSeen(m.hint);
  }

  bool unPeekEntryOmni(String shortLink) {
    //set this entry as seen
    ModeratorEntry? m = data.read(shortLink);
    if (m == null) return false;
    return _omni.unPeekEntry(m.hint);
  }

  bool peekThreadOmni(String shortLink) {
    //mark whole thread as seen
    return true;
  }

  bool dropEntryOmni(String shortLink) {
    //set this entry as seen
    return true;
  }

  bool dropThreadOmni(String shortLink) {
    //mark whole thread dropped
    return true;
  }

  Future<String> xxHashHex(Uint8List data) async {
    final attXX = data.hashCode;
    //final t = await xxHash(data).bake();
    final bdata = Uint8List(8)
      ..buffer.asByteData().setInt64(0, attXX, Endian.big);
    return HEX.encode(bdata);
  }

  List<ModeratorEntry>? updateLatestCapsule(
      Uint8List bin, bool stripModerateSeen) {
    //throw latest shit in the entrySet
    //TODO check for exact previous
    //error codes return less data
    //409,502
    if (bin.lengthInBytes < 28) return [];
    print('updateLatestCapsule len');
    print(bin.lengthInBytes);
    dataIn += bin.lengthInBytes;
    ModeratorCapsule mc;
    try {
      mc = ModeratorCapsule.fromBinary(bin);
    } catch (e) {
      //moderator caps outta sync
      return null;
    }
    //if (mc.isDEADBEEF) return null;
    if (mc == null) return null;
    if (mc.isEmpty) return null;

    if (mc.katamari.isNotEmpty) {
      _latestCapsule = mc;

      _disk?.commitKatamari(mc.katamari);
    }
    if (mc.threadEntries.isNotEmpty) {
      final t = insertToLocalEntrySet(mc, stripModerateSeen);
      if (t.isNotEmpty) {
        _dirty = true;
      }
      return t;
    }
    return null;
  }

  Future<List<ModeratorEntry>?> init(String _siteNamen, String server,
      String author, String avatar, interzoneDisk _disco) async {
    _server = server;
    _siteName = _siteNamen;
    _io = httpInterzone();

    _disk = _disco;
    //if we have incoming shit, show it
    _ogMessages
        .init(_disco, _siteName)
        .then((value) => currentPendingAsModeratorEntries);
    _omni.init(_disco, _siteName);

    //await _disk.init('interzoneData', false, _disk);
    String previousDartSeshId = '';
    // await _disk.readPreferences;

    /*List<String>? _prifs = await _disk?.readPreferences.then((_prifs) {
      if (_prifs != null) {
        previousDartSeshId = _prifs[0];
        if (_prifs.length > 1) {
          _nicke = _prifs[1];
          if (_prifs.length > 2) {
            _avatare = _prifs[2];
          }
        }
      } else {
        _nicke = author;
        _avatare = avatar;
      }
    }); */

    try {
      await _disk?.readIpfsPreferences.then((value) {
        if (value != null) {
          if (value.isNotEmpty) {
            final cd = value.split(':');
            _ipfsQ.setIpfsCredentials(cd[0], cd[1], '');
          }
        }
      });
    } catch (e) {
      print('world::readDeletesPreferences exploded');
      print(e);
    }

    try {
      _disk?.readDeletesPreferences.then((value) => {
            if (value != null)
              {
                if (value.isNotEmpty) {_deletedEntries = value.toSet()}
              }
          });
    } catch (e) {
      print('world::readDeletesPreferences exploded');
      print(e);
    }
    try {
      _disk?.readMutesPreferences.then((value) => {
            if (value != null)
              {
                if (value.isNotEmpty)
                  {
                    _mutedAuthors =
                        value.where((element) => element.isNotEmpty).toSet()
                  }
              }
          });
    } catch (e) {
      print('world::readMutesPreferences exploded');
      print(e);
    }

    _omni.init(_disk!, _siteName);

    await _xxCache.init(_disk!);
    _disk!.readCapsule.then((cps) {
      if (cps != null) {
        LocalEntrySetPrime(cps);
      }
    });

    /*Uint8List katt = await _disk!.readKatamari;
    if (katt.isNotEmpty) {
      _katamari = katt;
    }*/
    //async hundkarusellen!
    if (previousDartSeshId.isNotEmpty) {
      print('injected previous dart sesh id to help the police');
      _io!.dartSessionId = previousDartSeshId;
    }

    return login;
  }

  bool _refreshingViaHttp = false;
  Future<List<ModeratorEntry>?> refreshViaHTTP() async {
    if (_io == null) return null;
    if (_refreshingViaHttp) {
      return null;
    }
    if (_latestCapsule == null) return null;
    if (_latestCapsule!.katamari.isEmpty) return null;
    _refreshingViaHttp = true;
    //forget about failed requests
    //_xxCache.resetIPFSpullQueue();
    return await _io!
        .refresh(_server, _latestCapsule!.katamari)
        .then((hs) async {
      _refreshingViaHttp = false;
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        //await _disk.commitPreferences(_io!.dartSessionId, nick);
        satisfyXXwant(_io!.xxWantsList);
        updateLatestCapsule(hs, false);
        return outgoingMessageCacheSend(0).then((value) {
          if (value != null && value.isNotEmpty) {
            return value;
          }
          return null;
        });
      }
      /*if (_io?.online) {
        //TODO: request failed with timeout
      }*/
      return null;
    });
  }

  Future<List<ModeratorEntry>?> refreshTaggedViaHTTP(String shortLink) async {
    if (_io == null) return null;
    return await _io!.refreshTagged(_server, shortLink).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        return updateLatestCapsule(hs, false);
      }
    });
  }

  Future<List<ModeratorEntry>?> refreshAuthorViaHTTP(String shortLink) async {
    if (_io == null) return null;
    return await _io!.refreshAuthor(_server, shortLink).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        return updateLatestCapsule(hs, false);
      }
    });
  }

  Future<List<ModeratorEntry>?> refreshThreadViaHTTP(String shortlink) async {
    if (_io == null) return null;
    return await _io!.refreshThread(_server, shortlink).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        //await _disk.commitPreferences(_io!.dartSessionId, nick);
        satisfyXXwant(_io!.xxWantsList);
        return updateLatestCapsule(hs, false);
      }
      return null;
    });
  }

  Future<List<ModeratorEntry>?> refreshGroupThreadViaHTTP(
      String shortlink) async {
    if (_io == null) return null;
    return await _io!.refreshThread(_server, shortlink).then((hs) async {
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        //await _disk.commitPreferences(_io!.dartSessionId, nick);
        satisfyXXwant(_io!.xxWantsList);
        return updateLatestCapsule(hs, false);
      }
      return null;
    });
  }

  Future<bool> setUserNickAndImage(
      String _nick, String _avatar, Uint8List attachement) async {
    //await _disk?.commitPreferences(_io!.dartSessionId, _nick, _avatar);
    return true;
  }

  Future<bool> setIpfsCredentials(String k1, String k2, String url) async {
    //await _disk?.commitIpfsPreferences(k1 + ':' + k2);
    return _ipfsQ.setIpfsCredentials(k1, k2, url);
  }

  int get ipfsQflushFailed => _ipfsQ.flushFailed;
  Future<int> newThreadToQueue(
      String _baseUrl,
      String text,
      Uint8List attachement,
      String _nick,
      String _avatar,
      String shortLinkOfPrevious,
      String cid,
      String attachmentType,
      String xxHash,
      String _blurHash,
      int attachmentLength) async {
    if (shortLinkOfPrevious.isEmpty) {
      //patch in latest root
      final own = LatestThreads.latestMineShortLink;
      if (own.isEmpty) {
        shortLinkOfPrevious = LatestThreads.posts.last.shortLink;
      } else {
        shortLinkOfPrevious = own;
      }
    }

    final ogm = _ogMessages.addPost(
        _nick,
        _avatar,
        text,
        shortLinkOfPrevious,
        '',
        now,
        attachement,
        cid,
        attachmentType,
        xxHash,
        _blurHash,
        attachmentLength);

    if (ogm == null) return -1;

    //just push to disk when no net availablus
    _ogMessages.pushToDiskQueue(ogm.timestamp);

    //update cache of fake entries to show
    //dirty so flutter updates cache for queued entries
    _dirty = true;
    currentPendingAsModeratorEntries;
    return ogm.timestamp;
  }

  Map<int, int> _xxhc = {};
  Future<List<ModeratorEntry>?> newThread(
      String _baseUrl,
      String text,
      Uint8List attachement,
      String _nick,
      String _avatar,
      String shortLinkOfPrevious,
      String DEPcid,
      String mimeType,
      String _xxHash,
      String blurHash) async {
    if (_latestCapsule == null) return null;
    if (_latestCapsule!.katamari.isEmpty) {
      return null;
    }
    if (_nick.isEmpty) {
      _nick = currentNick;
    }
    if (shortLinkOfPrevious.isEmpty) {
      shortLinkOfPrevious =
          messages.all.where((element) => element.flags.isPost).first.shortLink;
    }
    Uint8List nt;
    //assume net is live here
    if (text.contains('https://')) {
      bool bcHash = text.contains('#bandcamp');
      bool tagBc = false;
      if (!text.startsWith('[')) {
        final ali = text.split('https://');
        final alib = ali[1].split(' ');
        final scresu = await scrapeBClink('https://' + alib[0]);
        if (scresu.isNotEmpty) {
          //pull the image as attachement

          if (scresu.last is Uint8List && attachement.isEmpty) {
            attachement = scresu.last;
            //patch url title pair
            text = '[' + scresu[0] + '](' + scresu[1] + ') ';
          } else {
            text = scresu[2];
          }

          if (!bcHash && scresu.contains('bandcamp')) tagBc = true;
          bool sk = false;
          alib.forEach((element) {
            if (!sk) {
              sk = true;
              return;
            }
            text += ' ' + element;
          });

          if (!bcHash && tagBc) text += ' #bandcamp';
        }
      }
    } //dont touch markdown links

    if (mimeType.isEmpty) {
      return await _io!
          .newThread(
              baseUrl,
              text,
              shortLinkOfPrevious,
              _latestCapsule != null ? _latestCapsule!.katamari : Uint8List(0),
              _nick,
              _avatar,
              Uint8List(0))
          .then((value) {
        if (value.isEmpty) return null;
        final na = updateLatestCapsule(value, false);
        if (na != null && na.isNotEmpty) {
          _dirty = true;
          currentPendingAsModeratorEntries;
        }
        return na;
      });
    }

    var attParamsToPass = Uint8List(0);
    if (_xxHash.isNotEmpty) {
      final atty = await _xxCache.pullLocalXX(_xxHash.toString());
      if (atty != null && atty.isNotEmpty) {
        final cid = await _xxCache.CIDforXX(_xxHash);
        if (cid.isEmpty) {
          //dont submit without ipfs
          //retry pushing
          _ipfsQ.pushToPinataIPFS('', '', atty).then((_cid) {
            if (_cid.isEmpty) {
              print('pushAttachment:pushToPinataIPFS failed');
              return;
            }
            xxC.mapCIDtoXX(_cid, _xxHash);
          });
          return [];
        }
        final rds = redressTextAndAttachment(
            text, atty.lengthInBytes, _xxHash, mimeType, cid);
        if (rds.isEmpty) {
          return null;
        }
        text = rds[0];
        attParamsToPass = rds[1];
      } else {
        print('newThread reported xx attacment not found, deep trouble.');
        return null;
      }
      if (blurHash.isNotEmpty) {
        text += ' !|' + blurHash + ' ';
      }
    } else {
      //xxhash not galculated yet
      print('noxx');
    }
    /*try {
        final image = img.decodeImage(attachement);
        final blurHash = BlurHash.encode(image!, numCompX: 4, numCompY: 3);
        text += ' !|' + blurHash.hash + ' ';
        print('blurhashed');
      } catch (e) {}*/

    return await _io!
        .newThread(baseUrl, text, shortLinkOfPrevious, _latestCapsule!.katamari,
            _nick, _avatar, attParamsToPass)
        .then((value) {
      if (value.isEmpty) return null;
      final na = updateLatestCapsule(value, false);
      if (na != null && na.isNotEmpty) {
        _dirty = true;
        currentPendingAsModeratorEntries;
      }
      return na;
    });
  }

  List<dynamic> redressTextAndAttachment(
      String text, int atLen, String atXX, String mimeType, String Cid) {
    var spicedCid = '';
    //0 image,1 video,2 audio
    bool isAudio = false;
    bool isVideo = false;
    bool isJpeg = false;
    bool isPng = false;
    bool isGif = false;
    bool isImage = false;

    //var mty = await lookupMimeType('', headerBytes: attachement);
    int atType = 0;
    switch (mimeType) {
      case 'image/jpeg':
        isJpeg = true;
        isImage = true;
        break;
      case 'image/png':
        isPng = true;
        isImage = true;
        break;
      case 'image/gif':
        isGif = true;
        isImage = true;

        break;
      case 'video/mp4':
        isVideo = true;
        atType = 1;
        break;
      case 'audio/mpeg':
        isAudio = true;
        atType = 2;
        break;
    }

    if (mimeType.startsWith('video')) {
      isVideo = true;
      atType = 1;
    }
    if (mimeType.startsWith('audio')) {
      isAudio = true;
      atType = 2;
    }
    if (!isImage && !isAudio && !isVideo) {
      return [];
    }
    if (Cid.isEmpty) {
      if (atType == 0) spicedCid = ' !&*';
      if (atType == 1) spicedCid = ' !∞*';
      if (atType == 2) spicedCid = ' !£*';
    } else {
      if (atType == 0) spicedCid = ' !&' + Cid;
      if (atType == 1) spicedCid = ' !∞' + Cid;
      if (atType == 2) spicedCid = ' !£' + Cid;
    }
    final ModeratorCapsuleItemAttachment att =
        ModeratorCapsuleItemAttachment(atLen, atXX, Cid);

    return [text + spicedCid, att.binary];
  }

  ModeratorCapsule? newGroup(String text, Uint8List attachement) {}
  ModeratorCapsule? newGroupChild(String text, Uint8List attachement) {}

  Future<List<ModeratorEntry>?> newReply(
      String text,
      String nick,
      String avatar,
      String rootShortlink,
      String ppShortlink,
      Uint8List attachement,
      String mimeType,
      String _xxHash,
      String _blurHash) async {
    if (_latestCapsule == null) return null;
    Uint8List nt;
    if (text.contains('https://')) {
      bool bcHash = text.contains('#bandcamp');
      bool tagBc = false;
      if (!text.startsWith('[')) {
        final ali = text.split('https://');
        final alib = ali[1].split(' ');
        final scresu = await scrapeBClink('https://' + alib[0]);
        if (scresu.isNotEmpty) {
          //pull the image as attachement
          if (scresu.last is Uint8List && attachement.isEmpty) {
            attachement = scresu.last;
            //patch url title pair
            text = '[' + scresu[0] + '](' + scresu[1] + ') ';
          } else {
            text = scresu[2];
          }

          if (!bcHash && scresu.contains('bandcamp')) tagBc = true;
          bool sk = false;
          alib.forEach((element) {
            if (!sk) {
              sk = true;
              return;
            }
            text += ' ' + element;
          });

          if (!bcHash && tagBc) text += ' #bandcamp';
        }
      }
    } //dont touch markdown links

    if (mimeType.isEmpty) {
      nt = await _io!.newReply(
          baseUrl,
          text,
          rootShortlink,
          ppShortlink,
          _latestCapsule != null ? _latestCapsule!.katamari : Uint8List(0),
          nick,
          avatar,
          Uint8List(0));

      if (nt.isEmpty) return null;
      //return empty list with dup
      final u = updateLatestCapsule(nt, false);
      if (u != null && u.isEmpty) return [];

      _dirty = true;
      currentPendingAsModeratorEntries;
      return threadByShortLink(rootShortlink).toList();
    }

    var attParamsToPass = Uint8List(0);
    final atty = await _xxCache.pullLocalXX(_xxHash.toString());
    if (atty != null && atty.isNotEmpty) {
      final cid = await _xxCache.CIDforXX(_xxHash);
      if (cid.isEmpty) {
        //dont submit without ipfs
        return [];
      }
      final rds = redressTextAndAttachment(
          text, atty.lengthInBytes, _xxHash, mimeType, cid);
      if (rds.isEmpty) {
        return null;
      }
      text = rds[0];
      attParamsToPass = rds[1];
      if (_blurHash.isNotEmpty) {
        text += ' !|' + _blurHash + ' ';
      }
    }

    //text += ' !&' + cid + ' ';
    /*xxC.pushCidData(cid, attachement);
    try {
      final image = img.decodeImage(attachement);
      final blurHash = BlurHash.encode(image!, numCompX: 4, numCompY: 3);
      text += ' !|' + blurHash.hash + ' ';
    } catch (e) {}*/
    if (_latestCapsule == null) {
      return null;
    }
    return await _io!
        .newReply(baseUrl, text, rootShortlink, ppShortlink,
            _latestCapsule!.katamari, nick, avatar, attParamsToPass)
        .then((value) {
      if (value.isEmpty) return null;
      //gives empty list on dup
      _dirty = true;
      currentPendingAsModeratorEntries;
      return updateLatestCapsule(value, false);
    });
  }

  Future<int> newReplyToQueue(
      String text,
      String shortLinkOfRoot,
      String shortLinkOfPP,
      Uint8List attachement,
      String attachmentType,
      String nick,
      String avatar,
      String _xxHash,
      String _blurHash,
      int attachmentLength) async {
    final ogm = _ogMessages.addReply(
        nick,
        avatar,
        text,
        shortLinkOfRoot,
        shortLinkOfPP,
        now,
        attachement,
        '',
        attachmentType,
        _xxHash,
        _blurHash,
        attachmentLength);
    if (ogm == null) return -1;
    //if (!_online) {
    //just push to disk when no net availablus
    _ogMessages.pushToDiskQueue(ogm.timestamp);

    //}
    //dirty so flutter updates cache for queued entries
    _dirty = true;
    //update cache of fake entries to shows
    currentPendingAsModeratorEntries;
    return ogm.timestamp;
  }

  Future<Uint8List?> readLocalFileByFilename(String filename) async {
    String cp = Directory.current.path;
    try {
      var did = await File(Directory.current.path + '/testImages/' + filename)
          .readAsBytes();
      return did;
    } catch (e) {
      return null;
    }
  }

  List<ModeratorEntry> insertToLocalEntrySet(
      ModeratorCapsule c, bool stripModerateSeen) {
    List<ModeratorEntry> _nw = [];
    final es = ModeratorEntrySet.fromModeratorCapsule(c.threadEntries);

    if (es.all.isEmpty) return _nw;
    var i = 0;
    for (var element in es.all) {
      if (_mutedAuthors.contains(element.author)) {
        continue;
      }
      if (_deletedEntries.contains(element.keyHint)) {
        continue;
      }
      //remember is seen
      //print(element.hint);
      //print(element.likes);
      element.flags.isSeen = _omni.isSeen(element.hint);
      //strip incoming moderate flag on local syb
      if (stripModerateSeen) {
        element.flags.moderate = false;
      }
      if (data.add(element)) {
        _nw.add(element);
        i++;
      }
    }

    if (i == 0) return _nw; //no new entries

    Uint8List ul = ModeratorCapsuleHelper(data);
    if (ul.isNotEmpty) {
      _disk?.commitCapsule(ul);
    }
    return _nw;
  }

  LocalEntrySetPrime(Uint8List l) {
    final es = ModeratorEntrySet.fromModeratorCapsule(l);
    if (es.all.isEmpty) return;
    var i = 0;
    for (var ii in es.all) {
      if (_mutedAuthors.contains(ii.author)) continue;
      if (_deletedEntries.contains(ii.keyHint)) continue;
      if (data.add(ii)) {
        if (ii.hasAttachment) {
          if (ii.ipfsCid.isNotEmpty) {
            xxC.mapCIDtoXX(ii.ipfsCid, ii.attachmentLinkAsXXint);
          }
        }
        i++;
      }
    }
    if (i > 0) _dirty = true;
    print('LocalEntrySetPrime with ' + i.toString());
  }

  Uint8List ModeratorCapsuleHelper(ModeratorEntrySet mode) {
    List<ModeratorCapsuleItem> mcl = [];

    if (mode.all.isEmpty) {
      print('PANIC: moderator capsule helper, moderator entryset is empty');
      return Uint8List(0);
    }

    //pluggy backlinks

    //pluggy perpre

    //mangle

    for (var ee in mode.all) {
      mcl.add(ModeratorCapsuleItem(
          ee.flags.binary,
          0,
          ee.bl,
          ee.pp,
          ee.timeElapsedSincePostMs,
          ee.body.length,
          ee.bodyBytes,
          ee.shortLinkBytes.lengthInBytes,
          ee.shortLinkBytes,
          ee.keyHint,
          ee.avatar,
          ee.nick,
          ee.attachmentLink,
          0,
          ee.perceivedTimestamp,
          ee.likes,
          ee.shares,
          ee.views,
          ee.replycount));
    }

    var fppf = mcl.last;
    var fppfb = mcl.last.binary;
    ModeratorCapsuleItem popo = ModeratorCapsuleItem.fromBinary(fppfb);

    Uint8List enl = Uint8List(5064000);
    final maxCaps = enl.lengthInBytes * 0.8;
    var wi = 0;
    for (var element in mcl.reversed) {
      if (wi < maxCaps) {
        element.binary.forEach((element2) {
          enl[wi++] = element2;
        });
      }
    }

    return enl.sublist(0, wi);
  }

  List<List<dynamic>> _xxPullQueue = [];
  Set<int> _xxCheckedQueue = {};
  int _xxpqcou = 0;

  scanForCachedAttachements() async {
    //proto pull all attachemendos
    _xxpqcou = 0;

    if (_xxCheckedQueue.length < _xxPullQueue.length) return;
    if (data.all.isNotEmpty) {
      for (var element in data.all) {
        if (element.flags.attachements) {
          if (_xxCache.isXXcached(element.attachmentLinkAsXXint)) return;
          //print(element.shortLink);
          _xxPullQueue.add([element.attachmentLinkAsXXint, element.shortLink]);
        }
      }
      ;
    }
    if (_xxPullQueue.isEmpty) return;
  }

  primeCurrentViewWithCachedAttachments(ModeratorEntrySet m, int amount) {
    for (var ii = 0; ii < m.all.length; ii++) {
      if (!data.all[ii].flags.attachements) continue;
      if (_xxCache.isXXcached(m.all[ii].attachmentLinkAsXXint)) continue;
      _xxPullQueue.add([m.all[ii].attachmentLinkAsXXint, m.all[ii].shortLink]);
      amount--;
      if (amount < 1) break;
    }
  }

  pullForCachedAttachements() async {
    if (_xxPullQueue.isEmpty) return;

    Uint8List? xx = await _xxCache
        .pullXX(_xxPullQueue[_xxpqcou][0], _xxPullQueue[_xxpqcou][1], _server)
        .then((value) => null);
    _xxCheckedQueue.add(_xxpqcou);
    _xxpqcou++;
    if (xx == null) return;
    print(xx.lengthInBytes);
    _xxCheckedQueue.add(_xxpqcou);
    _xxpqcou++;
  }

  Future<Uint8List> satisfyXXwant(Uint8List _wants) async {
    if (_wants.isEmpty) return Uint8List(0);
    var s = 0;
    var b = new ByteData.view(_wants.buffer);
    var w = 0;
    while (w < _wants.lengthInBytes) {
      var xx = b.getUint64(w);
      if (_xxCache.isXXcached(xx.toString())) {
        s = xx;
        break;
      }
      w += 8;
    }
    if (s == 0) return Uint8List(0);
    Uint8List? data = await _xxCache.pullXX(s.toString(), '', '');
    if (data != null) {
      //return _io!.satisfyXXwant(data);
      return data;
    }
    return Uint8List(0);
  }

  /*Future<String> pushToPinataIPFS(
      Uint8List data, pinata_api_key, pinata_secret_api_key) async {
    //add your pinata.cloud keys to PinataApiKeys.dart
    if (!haste) {
      return _ipfsQ.insertToQueue(xxHash(data), data);
    }
    return _io!.pushToPinataIPFS(pinata_api_key, pinata_secret_api_key, data);
  }*/
  Future<List<ModeratorEntry>?> dropEntry(String shortLink) async {
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return [];
    if (target.flags.moderate) {
      data.drop(target.keyHint);
      await _io?.newDrop(shortLink, _latestCapsule!.katamari);
      xxC.dropCID(target.attachmentLinkAsXXint, target.ipfsCid);
    }
    _dirty = true;
    return messages.all;
  }

  List<String> get droppedAttachments => data.droppedAttachements;
  int get attachmentsWeightInBytes => data.attachmentsWeightInBytes;

  Future<List<ModeratorEntry>?> dropEntryToQueue(
      String shortLink, int xx) async {
    if (xx != 0) {
      _ogMessages.hardDrop(xx);
    }
    ModeratorEntry? target = findModeratorEntryByShortLink(shortLink);
    if (target == null) return [];
    if (_deletedEntries.contains(target.keyHint)) return [];
    _deletedEntries.add(target.keyHint);
    //TODO kick off local storage too
    //xxC.delete(m.attachmentLink)
    //if it's your / your group, kill from serv too
    //await _disk?.commitPreferences(_io!.dartSessionId, _nick);
    await _disk?.commitDeletesPreferences(_deletedEntries.toList());
    if (target.flags.moderate) {
      _ogMessages.addDrop(target.shortLink, now);
    }
    //tell moderator to drop this

    _dirty = true;
    return messages.all;
  }

  bool _processingoutgoingMessageCacheSend = false;

  outgoingMessageCacheSendTimerRefresh() {
    if (_processingoutgoingMessageCacheSend) return null;
    //if (!_online) return null;

    Timer(Duration(milliseconds: 312), () {
      outgoingMessageCacheSend;
    });
  }

  Future<List<ModeratorEntry>?> outgoingMessageCacheSend(int timestamp) async {
    //if (_processingoutgoingMessageCacheSend) return null;
    //if (!_online) return null;
    //dont worry about online offline
    //if this request works, we are online
    OutgoingMessage? pi;
    if (timestamp != 0) {
      //we need a specific item
      pi = await _ogMessages.pickPendingItem(timestamp);
      //brute force this
      _processingoutgoingMessageCacheSend = false;
    } else {
      pi = _ogMessages.pendingItem;
    }

    if (pi == null) {
      _processingoutgoingMessageCacheSend = false;
      return null;
    }
    _processingoutgoingMessageCacheSend = true;
    //reset cache entry in case like

    //if the request does not work, this prevents blocking the queue
    ModeratorEntry? target =
        findModeratorEntryByShortLink(pi.shortLinkOfPrevious);
    if (target == null) {
      _ogMessages.releasePendingItem(pi.timestamp);
      //linking to dead mentry,drop
      _ogMessages.hardDrop(pi.timestamp);
      _processingoutgoingMessageCacheSend = false;
      currentPendingAsModeratorEntries;
      outgoingMessageCacheSendTimerRefresh();
      return null;
    }
    _ogMessages.holdPendingItem(pi.timestamp);
    /*if (pi.attachement.isNotEmpty) {
      xxC.pushXXbytes(pi.xxHash, pi.attachement, false);
    }*/
    if (pi.type == 3) {
      final r = await dropEntry(target.shortLink);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        //update cache of fake entries to shows
        currentPendingAsModeratorEntries;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 4) {
      final r = await muteEntry(target);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        //update cache of fake entries to shows
        currentPendingAsModeratorEntries;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 5) {
      final r = await shareEntry(target);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 6) {
      final r = await likeEntry(target);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _ogMessages.hardDrop(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    if (pi.type == 7) {
      final r = await unlikeEntry(target);

      if (r != null) {
        _ogMessages.commitSent(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        outgoingMessageCacheSendTimerRefresh();
        return r;
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    //null gives error, empty array a retry

    if (pi.type == 1) {
      final r = await newReply(
          pi.text,
          pi.nick,
          pi.avatar,
          pi.shortLinkOfPrevious,
          pi.shortLinkOfPP,
          pi.attachement,
          pi.mimeType,
          pi.xxHash,
          pi.blurHash);

      if (r != null) {
        if (r.isNotEmpty) {
          _ogMessages.commitSent(pi.timestamp);
          _ogMessages.hardDrop(pi.timestamp);
          _processingoutgoingMessageCacheSend = false;
          _dirty = true;
          //update cache of fake entries to shows
          currentPendingAsModeratorEntries;
          outgoingMessageCacheSendTimerRefresh();
          return r;
        }
      } else {
        //TODO reply timesouts
        _ogMessages.hardDrop(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        //_dirty = true;
        //update cache of fake entries to shows
        currentPendingAsModeratorEntries;
        _ogMessages.releasePendingItem(pi.timestamp);
        return [];
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }

    if (pi.type == 0 && !pi.hasAttachment) {
      final r = await newThread(
          baseUrl,
          pi.text,
          pi.attachement,
          pi.nick,
          pi.avatar,
          pi.shortLinkOfPrevious,
          pi.cid,
          pi.mimeType,
          pi.xxHash,
          pi.blurHash);
      if (r != null) {
        if (r.isNotEmpty) {
          _ogMessages.commitSent(pi.timestamp);
          _ogMessages.hardDrop(pi.timestamp);
          _processingoutgoingMessageCacheSend = false;
          _dirty = true;
          //update cache of fake entries to shows
          currentPendingAsModeratorEntries;
          outgoingMessageCacheSendTimerRefresh();
          return r;
        }
      } else {
        //_ogMessages.hardDrop(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        //_dirty = true;
        //update cache of fake entries to shows
        currentPendingAsModeratorEntries;
        _ogMessages.releasePendingItem(pi.timestamp);
        return [];
      }
      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }

    if (pi.type == 0 && pi.hasAttachment) {
      //inject pinata keys here

      final r = await newThread(
          baseUrl,
          pi.text,
          pi.attachement,
          pi.nick,
          pi.avatar,
          pi.shortLinkOfPrevious,
          pi.cid,
          pi.mimeType,
          pi.xxHash,
          pi.blurHash);
      if (r != null) {
        if (r.isNotEmpty) {
          _ogMessages.commitSent(pi.timestamp);
          _ogMessages.hardDrop(pi.timestamp);
          _processingoutgoingMessageCacheSend = false;
          _dirty = true;
          //update cache of fake entries to shows
          currentPendingAsModeratorEntries;
          outgoingMessageCacheSendTimerRefresh();
          return r;
        }
      } else {
        //_ogMessages.hardDrop(pi.timestamp);
        _processingoutgoingMessageCacheSend = false;
        _dirty = true;
        //update cache of fake entries to shows
        currentPendingAsModeratorEntries;
        _ogMessages.releasePendingItem(pi.timestamp);
        return [];
      }

      //we are offline, save on disk and try again layter
      _ogMessages.pushToDiskQueue(pi.timestamp);
      _ogMessages.releasePendingItem(pi.timestamp);
    }
    _processingoutgoingMessageCacheSend = false;
  }

  Future<List<ModeratorEntry>?> muteEntry(ModeratorEntry target) async {
    Uint8List nt;
    nt = await _io!.newMute(target.shortLink, _latestCapsule!.katamari);

    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt, false);
  }

  Future<int> muteAuthorToQueue(ModeratorEntry target) async {
    //if (m.flags.moderate) return false;
    if (_mutedAuthors.contains(target.keyHint)) return -1;
    _mutedAuthors.add(target.author);

    await _disk?.commitMutesPreferences(_mutedAuthors.toList());
    _dirty = true;
    final om = _ogMessages.addMute(target.shortLink, now);
    if (om == null) return -1;
    return om.timestamp;
  }

  Future<List<ModeratorEntry>?> likeEntry(ModeratorEntry target) async {
    Uint8List nt;
    //if (_omni.isLiked(target.hint)) return null;
    nt = await _io!.newLike(target.shortLink, _latestCapsule!.katamari);
    if (nt.isEmpty) return null;
    _omni.markLiked(target.hint);
    _dirty = true;
    return updateLatestCapsule(nt, false);
  }

  Future<int> likeEntryToQueue(ModeratorEntry target) async {
    Uint8List nt;
    if (_omni.isLiked(target.hint)) return -1;
    _omni.markLiked(target.hint);
    _dirty = true;
    final om = _ogMessages.addLike(target.shortLink, now);
    if (om == null) return -1;
    return om.timestamp;
  }

  Future<List<ModeratorEntry>?> unlikeEntry(ModeratorEntry target) async {
    Uint8List nt;
    if (_omni.isLiked(target.hint)) return null;
    nt = await _io!.newLike(target.shortLink, _latestCapsule!.katamari);
    if (nt.isEmpty) return null;
    _omni.unmarkLiked(target.hint);
    return updateLatestCapsule(nt, false);
  }

  Future<int> unlikeEntryToQueue(ModeratorEntry target) async {
    Uint8List nt;
    if (!_omni.isLiked(target.hint)) return -1;
    _omni.unmarkLiked(target.hint);
    _dirty = true;
    final om = _ogMessages.addunLike(target.shortLink, now);
    if (om == null) return -1;

    return om.timestamp;
  }

  Future<List<ModeratorEntry>?> shareEntry(ModeratorEntry target) async {
    Uint8List nt;
    nt = await _io!.newShare(target.shortLink, _latestCapsule!.katamari);
    _ogMessages.addShare(target.shortLink, now);
    if (nt.isEmpty) return null;
    return updateLatestCapsule(nt, false);
  }

  Future<int> shareEntryToQueue(ModeratorEntry target) async {
    final om = _ogMessages.addShare(target.shortLink, now);
    if (om == null) return -1;
    return om.timestamp;
  }

  ModeratorEntry? findModeratorEntryByShortLink(String shortLink) {
    return data.read(shortLink);
  }

  List<int> get stats {
    final s = [
      dataIn, //IOreceived
      dataOut, //IOsent
      dataPulledDisk, //disk pulled, total diskIO
      0, //memUsed
      0 //audience
    ];
    return s;
  }

  bool _loggingIn = false;
  Future<List<ModeratorEntry>?> get login async {
    //TODO init stuff
    if (_loggingIn) return null;
    _loggingIn = true;
    return _io!.handshake(_server, Uint8List(0)).then((hs) async {
      _loggingIn = false;
      if (hs.isNotEmpty) {
        dataIn += hs.lengthInBytes;
        //save prefs on sites layer not here
        //await _disk?.commitPreferences(
        //    _io!.dartSessionId, currentNick, currentAvatar);

        satisfyXXwant(_io!.xxWantsList);
        //final haa = ModeratorCapsule.fromBinary(hs);
        //send the pending stuff if we got some
        //refreshViaHTTP();
        return updateLatestCapsule(hs, false);
      }
      return data.all;
    });
    //return data.all;
  }

  Future<bool> get logout async {
    //TODO purge caches, send pending requests, etc

    return true;
  }

  Uint8List get outgoingCapsule {
    Uint8List occ = Uint8List(0);

    List<String> xxwants = [];
    //asks for missing stuff with a snapshot of my stuff

    List<ModeratorCapsuleItem> _mcl = [];
    int keepBel64kb = 0;
    for (ModeratorEntry ee in data.all) {
      if (_latelySeenp4p.contains(ee.keyHint)) continue;
      if (ee.flags.attachements) {
        if (!_xxCache.isXXstored(ee.attachmentLinkAsXXint)) {
          xxwants.add(ee.attachmentLinkAsXXint);
        }
        //capsules explode with too many entries
        if (keepBel64kb > 63980) break;
        _mcl.add(ModeratorCapsuleItem(
            ee.flags.binary,
            0,
            ee.bl,
            ee.pp,
            ee.timeElapsedSincePostMs,
            ee.body.length,
            ee.bodyBytes,
            ee.shortLinkBytes.lengthInBytes,
            ee.shortLinkBytes,
            ee.keyHint,
            ee.avatar,
            ee.nick,
            ee.attachmentLink,
            0,
            ee.perceivedTimestamp,
            ee.likes,
            ee.shares,
            ee.views,
            ee.replycount));

        keepBel64kb += _mcl.last.binary.lengthInBytes;
      }
    }
    Uint8List _l = Uint8List(128);
    var b0 = new ByteData.view(_l.buffer);

    xxwants.shuffle();
    /*int w = 0;
    for (var id in xxwants) {
      if (_latelySeenp4p.contains(id)) continue;
      if (w == _l.lengthInBytes) break;
      b0.setUint64(w, id);
      w += 8;
    }*/
    //IGNORE xxwants now

    Uint8List enl = Uint8List(66000);
    var wi = 0;
    for (var element in _mcl) {
      if (wi > 64000) {
        break;
      }
      for (var ib in element.binary) {
        enl[wi++] = ib;
      }
    }
    var w = 0;
    final ModeratorCapsule cap = ModeratorCapsule(Uint8List(0),
        enl.sublist(0, wi), Uint8List(0), _l.sublist(0, w), Uint8List(0));
    final zut = cap.binary;
    //final md = ModeratorCapsule.fromBinary(zut);
    return zut;
  }

  Set<int> _latelySeenp4p = {};
  bool _icpo = false;
  Future<Uint8List> incomingCapsule(Uint8List cap) async {
    if (_icpo) return Uint8List(0);

    //reply with xx I have. thats it.
    if (cap.lengthInBytes < 20) return outgoingCapsule;

    //inject entries and save, return list of new entries to you
    //also maps cid xxhash on diks
    //pass true so we strip moderate flag
    final _mismes = updateLatestCapsule(cap, true);

    //fullfill wants
    final passedCapsule = ModeratorCapsule.fromBinary(cap);
    if (passedCapsule.isEmpty) return Uint8List(0);
    _icpo = true;
    Uint8List satWanted = Uint8List(80000000);
    var b0 = new ByteData.view(satWanted.buffer);

    int sawa = 0;
    if (passedCapsule.xxWants.isNotEmpty) {
      List<int> wl = [];

      var mb = new ByteData.view(passedCapsule.xxWants.buffer);
      var wb = 0;
      while (wb < passedCapsule.xxWants.lengthInBytes) {
        wl.add(mb.getInt64(wb));
        wb += 8;
      }
/*
      for (var xxInt in wl) {
        final xs = await _xxCache.pullLocalXX(xxInt);
        if (xs != null) {
          if ((sawa + xs.lengthInBytes) < b0.lengthInBytes) {
            b0.setUint64(sawa, xs.lengthInBytes);
            sawa += 8;
            b0.setUint64(sawa, xxInt);
            sawa += 8;
            for (var xb in xs) {
              b0.setUint8(sawa++, xb);
            }
          }
        }
      } */
    }
    //inject new to me
    Set<int> _newt = {};
    Set<int> _receivedThreads = {};
    List<int> myxxwants = [];
/*
    if (passedCapsule.threadEntries.isNotEmpty) {
      final md =
          ModeratorEntrySet.fromModeratorCapsule(passedCapsule.threadEntries);

      for (var i in md.all) {
        _receivedThreads.add(i.keyHint);

        if (data.add(i)) {
          if (i.flags.attachements &&
              !_xxCache.isXXstored(i.attachmentLinkAsXXint)) {
            myxxwants.add(i.attachmentLinkAsXXint);
          }
          _newt.add(i.keyHint);
        }
      }
    }
*/
    //messages they might be missing
    //final List<ModeratorEntry> _mismes =
    //    data.all.where((element) => !_newt.contains(element.keyHint)).toList();

    List<ModeratorCapsuleItem> _mcl = [];
    var keepBel64kb = 0;
    List<int> satXXs = [];
    /*
    for (ModeratorEntry ee in _mismes) {
      if (keepBel64kb > 65000) break;
      if (ee.flags.attachements) {
        if (!_xxCache.isXXstored(ee.attachmentLinkAsXXint)) {
          myxxwants.add(ee.attachmentLinkAsXXint);
        }
        /*if (sawa == 0) {
          final xs =
              await _xxCache.pullIPFScid(ee.ipfsCid, ee.attachmentLinkAsXXint);
          if (xs != null) {
            b0.setUint64(sawa, xs.lengthInBytes);
            sawa += 8;
            b0.setUint64(sawa, ee.attachmentLinkAsXXint);
            sawa += 8;
            for (var xb in xs) {
              b0.setUint8(sawa++, xb);
            }
          }
        }*/
      }
      _mcl.add(ModeratorCapsuleItem(
          ee.flags.binary,
          0,
          ee.bl,
          ee.pp,
          ee.timeElapsedSincePostMs,
          ee.body.length,
          ee.bodyBytes,
          ee.shortLinkBytes.lengthInBytes,
          ee.shortLinkBytes,
          ee.keyHint,
          ee.avatar,
          ee.nick,
          ee.attachmentLink,
          0,
          ee.perceivedTimestamp,
          ee.likes,
          ee.shares,
          ee.views,
          ee.replycount));
      keepBel64kb += _mcl.last.binary.lengthInBytes;
    }
    */
    //only 64 kb a pop
    Uint8List enl = Uint8List(68000);
    var wic = 0;
    for (var element in _mcl) {
      //do not send stuff in recent circulation
      if (_latelySeenp4p.contains(element.hint)) continue;
      _latelySeenp4p.add(element.hint);

      if (wic > 64000) {
        break;
      }
      for (var ib in element.binary) {
        enl[wic++] = ib;
      }
    }
    //get the xxhashes I'm missing

    /*if (sawa == 0 && wic == 0 && myxxwants.isEmpty) {
      //no xxs to satisfy request, no new messages
      _icpo = false;
      return Uint8List(0);
    }*/

    Uint8List _xxWantBin = Uint8List(8089);
    var bo = new ByteData.view(_xxWantBin.buffer);
    myxxwants.shuffle();
    int xw = 0;
    for (var id in myxxwants) {
      if (_latelySeenp4p.contains(id)) continue;
      if (xw == _xxWantBin.lengthInBytes) break;
      bo.setUint64(xw, id);
      xw += 8;
    }

    //TODO something missing for sender?

    final ccap = ModeratorCapsule(Uint8List(0), enl.sublist(0, wic),
        Uint8List(0), _xxWantBin.sublist(0, xw), satWanted.sublist(0, sawa));
    final zut = ccap.binary;
    final md = ModeratorCapsule.fromBinary(zut);
    _icpo = false;
    return zut;
  }

  Future<Uint8List> outgoingCapsuleXXPush(String xxInt) async {
    Uint8List satWanted = Uint8List(80000000);
    var b0 = new ByteData.view(satWanted.buffer);

    int sawa = 0;

    final xs = await _xxCache.pullLocalXX(xxInt.toString());
    if (xs != null) {
      if ((sawa + xs.lengthInBytes) < b0.lengthInBytes) {
        b0.setUint64(sawa, xs.lengthInBytes);
        sawa += 8;
        b0.setUint64(sawa, 0); //UNFUK to xx
        sawa += 8;
        for (var xb in xs) {
          b0.setUint8(sawa++, xb);
        }
      }
    }
    if (sawa == 0) return Uint8List(0);
    final ccap = ModeratorCapsule(Uint8List(0), Uint8List(0), Uint8List(0),
        Uint8List(0), satWanted.sublist(0, sawa));
    final zut = ccap.binary;
    //final md = ModeratorCapsule.fromBinary(zut);
    _icpo = false;
    return zut;
  }

  Future<Uint8List> incomingCapsuleXXdata(Uint8List cap) async {
    //reply with xx I have. thats it.

    //fullfill wants
    final passedCapsule = ModeratorCapsule.fromBinary(cap);
    if (passedCapsule.isEmpty) return Uint8List(0);

    List<int> satXXs = [];
    Set<int> _newt = {};
    Set<int> _receivedThreads = {};
    List<String> myxxwants = [];

    if (passedCapsule.threadEntries.isNotEmpty) {
      final md =
          ModeratorEntrySet.fromModeratorCapsule(passedCapsule.threadEntries);

      for (var i in md.all) {
        if (_latelySeenp4p.contains(i.keyHint)) continue;
        _latelySeenp4p.add(i.keyHint);

        _receivedThreads.add(i.keyHint);

        if (data.add(i)) {
          if (i.flags.attachements &&
              !_xxCache.isXXstored(i.attachmentLinkAsXXint)) {
            myxxwants.add(i.attachmentLinkAsXXint);
          }
          _newt.add(i.keyHint);
        }
      }
    }

    if (passedCapsule.xxData.isNotEmpty) {
      //satisfy xxWants one by one
      /*
      int idx = 0;
      while (idx < passedCapsule.xxData.lengthInBytes) {
        var x0 = new ByteData.view(passedCapsule.xxData.buffer);
        int le = x0.getUint64(idx);
        int xx = x0.getUint64(idx + 8);
        _latelySeenp4p.add(xx);

        idx += 16;
//pushXXbytes
        final sca = await _xxCache.satisfy(
            xx, passedCapsule.xxData.sublist(idx, idx + le), false);
        if (sca) {
          satXXs.add(xx);
        }
        idx += le;
      } */
    }

    return Uint8List(0);
  }

  Uint8List outgoingCapsuleXXwant(int xxHash) {
    Uint8List occ = Uint8List(0);
    //dont ask for things already in
    //if (_latelySeenp4p.contains(xxHash)) return occ;
    List<int> xxwants = [];
    xxwants.add(xxHash);
    //asks for missing stuff with a snapshot of my stuff

    List<ModeratorCapsuleItem> _mcl = [];

    Uint8List _l = Uint8List(128);
    var b0 = new ByteData.view(_l.buffer);

    int w = 0;
    for (var id in xxwants) {
      if (w == _l.lengthInBytes) break;
      b0.setUint64(w, id);
      w += 8;
    }

    final ModeratorCapsule cap = ModeratorCapsule(Uint8List(0), Uint8List(0),
        Uint8List(0), _l.sublist(0, w), Uint8List(0));
    final zut = cap.binary;
    return zut;
  }

  Map<String, String> _blurHashIndex = {};
  Future<String> pushAttachment(Uint8List data, String? mimeType,
      String _xxHashx, String _blurHash) async {
    //return -1 on fail
    if (data.isEmpty) return '';
    //normalize hex
    String _xxHash = _xxHashx.toLowerCase();
    //assume jpeg for now TODO
    _blurHashIndex[_xxHash] = _blurHash;
    print('pushAttachment:' + _xxHash.toString());
    if (_xxHash.isNotEmpty) {
      return xxC.pushXXbytes(_xxHash, data, false).then((value2) async {
        if (!value2) {
          print('pushAttachment:xxC.pushXXbytes failed' + value2.toString());

          return '';
        }
        print('pushing ipfs attachemnt len ' + data.lengthInBytes.toString());
        _ipfsQ.pushToPinataIPFS('', '', data).then((_cid) {
          if (_cid.isEmpty) {
            print('pushAttachment:pushToPinataIPFS failed');

            return _xxHash;
          }
          xxC.mapCIDtoXX(_cid, _xxHash);
        });

        return _xxHash;
      });
    }
    return xxHash(data).bake().then((value) async {
      return xxC.pushXXbytes(value.toLowerCase(), data, false).then((value2) {
        if (!value2) return '';
        print('attachemnt len ' + data.lengthInBytes.toString());
        print(value);
        _ipfsQ.pushToPinataIPFS('', '', data).then((_cid) {
          if (_cid.isEmpty) return value.toLowerCase();
          xxC.mapCIDtoXX(_cid, value.toLowerCase());
        });
        return value.toLowerCase();
      });
    });
  }
}

class compressImageIsolate {
  compressImageIsolate(this._attachementData);
  final List<int> _attachementData;
  Future<Uint8List> bake() async {
    final p = ReceivePort();
    // spawn the isolate and wait for it to complete
    await Isolate.spawn(_compressImageIsolate, p.sendPort);
    return await p.first;
  }

  Future<void> _compressImageIsolate(SendPort p) async {
    final dad = Uint8List.fromList(_attachementData);
    final input = ImageFile(
      rawBytes: dad,
      filePath: 'gaggi',
    );
    final output = compress(ImageFileConfiguration(
        input: input,
        config: Configuration(jpgQuality: 52, outputType: OutputType.jpg)));

    Isolate.exit(p, output.rawBytes);
  }
}

class xxHash {
  xxHash(this._xxData);
  final List<int> _xxData;

  Future<String> bake() async {
    final p = ReceivePort();
    // spawn the isolate and wait for it to complete
    await Isolate.spawn(_xxHashIsolate, p.sendPort);
    // get and return the result data
    return await p.first;
  }

  Future<void> _xxHashIsolate(SendPort p) async {
    //digest = _xxData.hashCode;
    final digest = xxHash64.convert(_xxData);
    Isolate.exit(p, digest.toString());
  }
}

Uint8List binString(String s) {
  if (s.isEmpty) return Uint8List(2);
  Uint8List b = Uint8List(1024);
  var bdata = new ByteData.view(b.buffer);

  List<int> aa = utf8.encode(s);
  bdata.setUint16(0, aa.length);
  int wi = 2;
  for (var i in aa) {
    bdata.setUint8(wi++, i);
  }
  return b.sublist(0, wi);
}
